/*  Screem:  screem-tag_file.c
 *
 *  Copyright (C) 2003 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include "config.h"

#include "screem-tagfile.h"

#include <string.h>
#include <libgnome/gnome-macros.h>
#include <libgnome/gnome-i18n.h>
#include <gtk/gtktreestore.h>
#include <gtk/gtktreemodelsort.h>
#include <glib/gmarkup.h>
#include <gdk/gdk.h>

#include "fileops.h"
#include "screem-file-browser.h"

static void screem_tag_file_class_init( ScreemTagFileClass *klass );
static void screem_tag_file_instance_init( ScreemTagFile *tag_file );
static void screem_tag_file_finalize( GObject *object );
static void screem_tag_file_set_prop( GObject *object, guint prop_id,
				const GValue *value, GParamSpec *spec );
static void screem_tag_file_get_prop( GObject *object, guint prop_id,
				GValue *value, GParamSpec *spec );

static gint screem_tag_file_compare_func( GtkTreeModel *model, 
				    	  GtkTreeIter *a, 
					  GtkTreeIter *b,
				    	  gpointer data );
static void screem_tag_file_icon_change( ScreemFileBrowser *browser, 
					const gchar *uri,
					const gchar *mime_type,
					GtkTreeIter *it,
					gpointer data );
static void screem_tag_file_free_node_info( NodeInfo *info );
static void screem_tag_file_clear( ScreemTagFile *tagfile );
static gboolean screem_tag_file_clear_foreach( GtkTreeModel *model,
						GtkTreePath *path,
						GtkTreeIter *iter,
						gpointer data );
static gboolean screem_tag_file_clear_paths( gpointer key, 
						gpointer value,
						gpointer data );

static void screem_tag_file_start_element( GMarkupParseContext *context,
					const gchar *element,
					const gchar **attributes,
					const gchar **values,
					gpointer data,
					GError **error );
static void screem_tag_file_end_element( GMarkupParseContext *context,
					const gchar *element,
					gpointer data,
					GError **error );
static void screem_tag_file_text( GMarkupParseContext *context,
				const gchar *text,
				gsize len,
				gpointer data,
				GError **error );
static void screem_tag_file_error( GMarkupParseContext *context,
				GError *error,
				gpointer data );
static void screem_tag_file_parse_node( ScreemTagFilePrivate *priv,
					const gchar *element,
					const gchar **attributes,
					const gchar **values,
					gboolean group );
static void screem_tag_file_parse_insert( ScreemTagFilePrivate *priv,
					  const gchar **attributes,
					  const gchar **values );

static void screem_tag_file_parse_param( ScreemTagFilePrivate *priv,
					 const gchar **attributes,
					 const gchar **values );

enum {
	PROP_0,
	PROP_ACTIVE
};

struct ScreemTagFilePrivate {
	ScreemTagFile *tfile;
	
	gchar *filename;

	gchar *name;
	gchar *description;

	gchar *type;
	
	ScreemFileBrowser *browser;
	/* the unsorted browser model */
	GtkTreeModel *model;
	
	/* used when parsing */
	GSList *states;
	GSList *iters;
	GHashTable *paths;

	gboolean active;

	gboolean loaded;

	/* current auto complete prefix */
	const gchar *prefix;
};

static GMarkupParser parser_cbs = {
	screem_tag_file_start_element,
	screem_tag_file_end_element,
	screem_tag_file_text,
	NULL,
	screem_tag_file_error
};

typedef enum {
	PARSER_STATE_NONE = 0,
	PARSER_STATE_REF,
	PARSER_STATE_GROUP,
	PARSER_STATE_ENTRY,
	PARSER_STATE_FUNCTION,
	PARSER_STATE_INSERT,
	PARSER_STATE_DESCRIPTION,
	PARSER_STATE_PARAM,
	PARSER_STATE_ATTRIBUTE
} ScreemTagTreeParserState;

static const gchar *parser_tags[] = {
	"",
	"ref",
	"group",
	"entry",
	"function",
	"insert",
	"description",
	"param",
	"property",
};


GNOME_CLASS_BOILERPLATE( ScreemTagFile, screem_tag_file,
			  GObject,
			  G_TYPE_OBJECT )

static void screem_tag_file_class_init( ScreemTagFileClass *klass )
{
	GObjectClass *obj_class;
	GParamSpec *spec;

	obj_class = G_OBJECT_CLASS( klass );
	obj_class->finalize = screem_tag_file_finalize;
	obj_class->get_property = screem_tag_file_get_prop;
	obj_class->set_property = screem_tag_file_set_prop;

	spec = g_param_spec_boolean( "active", "active",
				_( "Is the Tag File active or no" ),
				TRUE, 
				G_PARAM_READABLE | G_PARAM_WRITABLE );
	g_object_class_install_property( obj_class, PROP_ACTIVE,
					spec );
}

static void screem_tag_file_instance_init( ScreemTagFile *tag_file )
{
	ScreemTagFilePrivate *priv;
	
	priv = tag_file->priv = g_new0( ScreemTagFilePrivate, 1 );
	priv->tfile = tag_file;

	priv->browser = screem_file_browser_new();
	priv->model = screem_file_browser_get_model( priv->browser );
	priv->model = gtk_tree_model_sort_get_model( GTK_TREE_MODEL_SORT( priv->model ) );

	g_object_set_data( G_OBJECT( priv->model ), "tagfile", tag_file );
	
	screem_file_browser_set_mode( priv->browser,
					FILE_BROWSE_RECURSE );
	screem_file_browser_set_sort_func( priv->browser,
					   screem_tag_file_compare_func,
					   NULL );
	priv->paths = g_hash_table_new( g_str_hash, g_str_equal );
	
	g_signal_connect( G_OBJECT( priv->browser ),
			  "icon_change", 
			  G_CALLBACK( screem_tag_file_icon_change ),
			  NULL );

}

static void screem_tag_file_finalize( GObject *object )
{
	ScreemTagFile *tag_file;
	ScreemTagFilePrivate *priv;
	
	g_return_if_fail( object != NULL );
	g_return_if_fail( SCREEM_IS_TAGFILE( object ) );

	tag_file = SCREEM_TAGFILE( object );

	priv = tag_file->priv;

	screem_tag_file_clear( tag_file );

	g_hash_table_destroy( priv->paths );
	
	g_free( priv->filename );
	
	g_object_unref( priv->browser );
	
	g_free( priv );
	
	GNOME_CALL_PARENT( G_OBJECT_CLASS, finalize, (object) );
}

static void screem_tag_file_set_prop( GObject *object, guint prop_id,
				const GValue *value, GParamSpec *spec )
{
	ScreemTagFile *tag_file;
	ScreemTagFilePrivate *priv;
	
	tag_file = SCREEM_TAGFILE( object );
	priv = tag_file->priv;

	switch( prop_id ) {
		case PROP_ACTIVE:
			priv->active = g_value_get_boolean( value );
			if( priv->active && ! priv->loaded ) {
				screem_tag_file_load( tag_file );
			}
			break;
		default:
			break;
	}
}

static void screem_tag_file_get_prop( GObject *object, guint prop_id,
				GValue *value, GParamSpec *spec )
{
	ScreemTagFile *tag_file;
	ScreemTagFilePrivate *priv;
	
	tag_file = SCREEM_TAGFILE( object );
	priv = tag_file->priv;

	switch( prop_id ) {
		case PROP_ACTIVE:
			g_value_set_boolean( value, priv->active );
			break;
		default:
			break;
	}
}

/* static stuff */
static gint screem_tag_file_compare_func( GtkTreeModel *model, 
				    	  GtkTreeIter *a, 
					  GtkTreeIter *b,
				    	  gpointer data )
{
        NodeInfo *ainfo;
        NodeInfo *binfo;
	int ret = 0;

	gtk_tree_model_get( model, a, 
			FILE_BROWSER_DATA_COL, &ainfo, -1 );
	gtk_tree_model_get( model, b, 
			FILE_BROWSER_DATA_COL, &binfo, -1 );

	if( ainfo && ! binfo ) {
		ret = 1;
	} else if( binfo && !ainfo ) {
		ret = -1;
	} else if( ! ainfo && ! binfo ) {
		ret = 0;
	} else {
		ret = strcmp( ainfo->name, binfo->name );
	}
	
	return ret;
}

static void screem_tag_file_icon_change( ScreemFileBrowser *browser, 
					const gchar *uri,
					const gchar *mime_type,
					GtkTreeIter *it,
					gpointer data )
{
	GtkTreeModel *model;
	NodeInfo *info;
	GdkPixbuf *pixbuf;
	
	model = screem_file_browser_get_model( browser );
	model = gtk_tree_model_sort_get_model( GTK_TREE_MODEL_SORT( model ) );

	gtk_tree_model_get( model, it, 
			FILE_BROWSER_DATA_COL, &info,
			-1 );
	if( info ) {
		pixbuf = NULL;
		if( info->file ) {
			pixbuf = screem_file_browser_get_icon( browser,
								info->file,
								16, 16, FALSE );
		} else if( gtk_tree_model_iter_has_child( model, it ) ) {
			pixbuf = screem_file_browser_get_icon( browser,
								"/",
								16, 16, FALSE );
		}
		gtk_tree_store_set( GTK_TREE_STORE( model ), 
				it, FILE_BROWSER_ICON_COL,
				pixbuf, -1 );
		if( pixbuf ) {
			g_object_unref( pixbuf );
		}
	}
}

static void screem_tag_file_free_node_info( NodeInfo *info )
{
	GSList *list;
	NodeAttribute *attr;
	
	g_free( info->name );
	g_free( info->open );
	g_free( info->close );
	g_free( info->file );
	g_free( info->text );
	g_free( info->description );

	for( list = info->attributes; list; list = list->next ) {
		attr = (NodeAttribute*)list->data;

		g_free( attr->name );
		g_free( attr->type );
		g_free( attr->defvalue );
		g_free( attr->description );
		g_free( attr );
	}
	if( info->attributes ) {
		g_slist_free( info->attributes );
	}
	
	g_free( info );
}

static void screem_tag_file_clear( ScreemTagFile *tagfile )
{
	ScreemTagFilePrivate *priv;
	
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );

	priv = tagfile->priv;
	
	gtk_tree_model_foreach( priv->model,
			screem_tag_file_clear_foreach,
			NULL );

	g_hash_table_foreach_steal( priv->paths,
			(GHRFunc)screem_tag_file_clear_paths,
			priv->name );

	gtk_tree_store_clear( GTK_TREE_STORE( priv->model ) );
}

static gboolean screem_tag_file_clear_foreach( GtkTreeModel *model,
						GtkTreePath *path,
						GtkTreeIter *iter,
						gpointer data )
{
	NodeInfo *info;
	
	gtk_tree_model_get( model, iter, 
			FILE_BROWSER_DATA_COL, &info, 
			-1 );
	if( info ) {
		screem_tag_file_free_node_info( info );
	}

	return FALSE;
}

static gboolean screem_tag_file_clear_paths( gpointer key, 
						gpointer value,
						gpointer data )
{
	const gchar *prefix;
	gboolean ret;

	prefix = (const gchar*)data;
	
	ret = FALSE;
	if( ! strncmp( prefix, key + 1, strlen( prefix ) ) ) {
		g_free( key );
		gtk_tree_path_free( value );
		ret = TRUE;
	}

	return ret;
}

static void screem_tag_file_start_element( GMarkupParseContext *context,
					const gchar *element,
					const gchar **attributes,
					const gchar **values,
					gpointer data,
					GError **error )
{
	ScreemTagFilePrivate *priv;
	ScreemTagTreeParserState state;

	priv = (ScreemTagFilePrivate*)data;
	state = GPOINTER_TO_INT( priv->states->data );
	
	switch( state ) {
		case PARSER_STATE_NONE:
			if( ! strcmp( "ref", element ) ) {
				screem_tag_file_parse_node( priv,
							    element,
							    attributes,
							    values, TRUE );
				state = PARSER_STATE_REF;
			} else {
				/* unknown root element, ignore */
			}
			break;
		case PARSER_STATE_REF:
			/* ref can't have entry / function according to our dtd,
			but it simplifies adding the current doctype branch,
			so we allow it here */
		case PARSER_STATE_GROUP:
			if( ! strcmp( "group", element ) ) {
				screem_tag_file_parse_node( priv,
							    element,
							    attributes,
							    values, TRUE );
				state = PARSER_STATE_GROUP;
			} else if( ! strcmp( "entry", element ) ) {
				screem_tag_file_parse_node( priv,
							    element,
							    attributes,
							    values, FALSE );
				state = PARSER_STATE_ENTRY;
			} else if( ! strcmp( "function", element ) ) {
				screem_tag_file_parse_node( priv,
							    element,
							    attributes,
							    values, FALSE );
				state = PARSER_STATE_FUNCTION;
			}
			break;
		case PARSER_STATE_ENTRY:
			if( ! strcmp( "insert", element ) ) {
				screem_tag_file_parse_insert( priv,
								attributes,
								values );
				state = PARSER_STATE_INSERT;
			} else if( ! strcmp( "description", element ) ) {
				state = PARSER_STATE_DESCRIPTION;
			} else if( ! strcmp( "property", element ) ) {
				screem_tag_file_parse_param( priv,
				 			attributes,
							values );
				state = PARSER_STATE_ATTRIBUTE;
			}
			break;
		case PARSER_STATE_FUNCTION:
			if( ! strcmp( "insert", element ) ) {
				screem_tag_file_parse_insert( priv,
								attributes,
								values );
				state = PARSER_STATE_INSERT;
			} else if( ! strcmp( "description", element ) ) {
				state = PARSER_STATE_DESCRIPTION;
			} else if( ! strcmp( "param", element ) ) {
				screem_tag_file_parse_param( priv,
				 			attributes,
							values );
				state = PARSER_STATE_PARAM;
			} else {
				state = PARSER_STATE_NONE;
			}
			break;
		case PARSER_STATE_INSERT:
		case PARSER_STATE_DESCRIPTION:
		case PARSER_STATE_PARAM:
		case PARSER_STATE_ATTRIBUTE:
			/* states should not occur */
			state = PARSER_STATE_NONE;
			break;
		default:
			/* unknown state, shouldn't happen */
			state = PARSER_STATE_NONE;
			break;	
	}

	if( state != PARSER_STATE_NONE ) {
		priv->states = g_slist_prepend( priv->states,
					GINT_TO_POINTER( state ) );
	}
}

static void screem_tag_file_end_element( GMarkupParseContext *context,
					const gchar *element,
					gpointer data,
					GError **error )
{
	ScreemTagFilePrivate *priv;
	ScreemTagTreeParserState state;
	GSList *top;
	NodeInfo *info;
	
	priv = (ScreemTagFilePrivate*)data;
	state = GPOINTER_TO_INT( priv->states->data );
	top = priv->states;

	switch( state ) {
		case PARSER_STATE_NONE:
			/* shouldn't happen */
			state = PARSER_STATE_NONE;
			break;
		default:
			if( strcmp( parser_tags[ state ], element ) ) {
				state = PARSER_STATE_NONE;
			}
			break;
	}

	if( state != PARSER_STATE_NONE ) {
		priv->states = top->next;
		top->next = NULL;
		g_slist_free( top );
		if( state == PARSER_STATE_REF ||
		    state == PARSER_STATE_GROUP ||
		    state == PARSER_STATE_ENTRY ||
		    state == PARSER_STATE_FUNCTION ) {
			top = priv->iters;
			priv->iters = top->next;
			top->next = NULL;
			if( top->data ) {
				info = NULL;
				gtk_tree_model_get( GTK_TREE_MODEL( priv->model ),
						  top->data, FILE_BROWSER_DATA_COL,
						  &info, -1 );
				if( info && info->attributes ) {
					info->attributes = g_slist_reverse( info->attributes );
				}
	
 				gtk_tree_iter_free( (GtkTreeIter*)top->data );
			}
			g_slist_free( top );
		}
	}
}

static void screem_tag_file_text( GMarkupParseContext *context,
				const gchar *text,
				gsize len,
				gpointer data,
				GError **error )
{
	ScreemTagFilePrivate *priv;
	ScreemTagTreeParserState state;
	GSList *top;
	GtkTreeIter *it;
	NodeInfo *info;
	
	priv = (ScreemTagFilePrivate*)data;
	state = GPOINTER_TO_INT( priv->states->data );
		
	top = priv->iters;
	
	it = top->data;
      	gdk_threads_enter();
	info = NULL;
	if( it ) {
		gtk_tree_model_get( GTK_TREE_MODEL( priv->model ),
					  it, FILE_BROWSER_DATA_COL,
					  &info, -1 );
	}	
     	gdk_threads_leave();
	
	switch( state ) {
		case PARSER_STATE_INSERT:
			if( info ) {
				info->text = g_strdup( text );
				info->text = g_strstrip( info->text );
			}
			break;
		case PARSER_STATE_DESCRIPTION:
			if( info ) {
				info->description = g_strdup( text );
				info->description = g_strstrip( info->description );
			}
			break;
		case PARSER_STATE_PARAM:
		case PARSER_STATE_ATTRIBUTE:
			if( info && info->attributes ) {
				NodeAttribute *attr;

				attr = (NodeAttribute*)info->attributes->data;

				if( attr ) {
					attr->description = g_strdup( text );
					attr->description = g_strstrip( attr->description );
				}
			}
			break;
		default:
			/* shouldn't happen */
			break;
	}
}

static void screem_tag_file_error( GMarkupParseContext *context,
				GError *error,
				gpointer data )
{
	
}

static void screem_tag_file_parse_node( ScreemTagFilePrivate *priv,
					const gchar *element,
					const gchar **attributes,
					const gchar **values,
					gboolean group )
{
	GSList *top;
	GtkTreeIter it;
	GtkTreeIter *iter;
	const gchar *name;
	const gchar *description;
	const gchar *type;
	gint i;
	GtkTreePath *path;
	NodeInfo *info;

	GString *fullpath;
	GSList *paths;
	
	top = priv->iters;
	
	info = NULL;	
	name = description = type = NULL;
	for( i = 0; attributes[ i ]; ++ i ) {
		if( ! strcmp( "name", attributes[ i ] ) ) {
			name = values[ i ];
		} else if( ! strcmp( "description", attributes[ i ] ) ) {
			description = values[ i ];
		} else if( ! strcmp( "type", attributes[ i ] ) ) {
			type = values[ i ];
		}
	}
	
	/* name="" is required, but we will create one if it doesn't exist */
	if( ! name ) {
		name = _( "Unknown Resource" );
	}

	/* set name/desc on the ScreemTagFile if this is the
	 * ref element we are parsing */
	if( ! strcmp( "ref", element ) ) {
		g_free( priv->name );
		g_free( priv->description );
		priv->name = g_strdup( name );
		priv->description = NULL;
		if( description ) {
			priv->description = g_strdup( description );
		}
		if( type ) {
			priv->type = g_strdup( type );
		}
	}
	
	/* see if we already have a node matching name at this level,
	   if so we just push that iter instead of creating a new one */

      	gdk_threads_enter();
	fullpath = g_string_new( name );
	for( paths = top; paths; paths = paths->next ) {
		GtkTreeIter *pit;
		
		pit = (GtkTreeIter*)paths->data;
		if( pit ) {
			gtk_tree_model_get( GTK_TREE_MODEL( priv->model ),
						  pit, FILE_BROWSER_DATA_COL,
						  &info, -1 );
			g_string_prepend_c( fullpath, '/' );
			g_string_prepend( fullpath, info->name );
		}
	}
	g_string_prepend_c( fullpath, '/' );
	path = g_hash_table_lookup( priv->paths, fullpath->str );

	if( path && ! gtk_tree_model_get_iter( GTK_TREE_MODEL( priv->model ),
						&it, path ) ) {
		path = NULL;
	} else if( path && ! strcmp( _( "Current Doctype" ), fullpath->str + 1 ) ) {
		gtk_tree_store_remove( GTK_TREE_STORE( priv->model ), &it );
		g_hash_table_foreach_steal( priv->paths,
				screem_tag_file_clear_paths,
				_( "Current Doctype" ) );
		path = NULL;
	}
	
	if( ! path ) {
		GdkPixbuf *pixbuf;
		
		gtk_tree_store_append( GTK_TREE_STORE( priv->model ),
					&it, (GtkTreeIter*)top->data );						 
		
		pixbuf = NULL;
		if( group ) {
			pixbuf = screem_file_browser_get_icon( priv->browser,
								"/", 16, 16,
								FALSE );
		}

		info = g_new0( NodeInfo, 1 );
		info->tfile = priv->tfile;
		info->name = g_strdup( name );
		if( description && group ) {
			info->description = g_strdup( description );
		}
		gtk_tree_store_set( GTK_TREE_STORE( priv->model ), 
					&it, 
					FILE_BROWSER_NAME_COL, name,
					FILE_BROWSER_ICON_COL, pixbuf,
					FILE_BROWSER_DATA_COL, info,
					-1 );
		if( pixbuf ) {
			g_object_unref( pixbuf );
		}
		path = gtk_tree_model_get_path( GTK_TREE_MODEL( priv->model ),
						&it );
		g_hash_table_insert( priv->paths, fullpath->str,
				     path );

		path = NULL;
	 }
	
	iter = gtk_tree_iter_copy( &it );
	priv->iters = g_slist_prepend( priv->iters, iter );
	
	g_string_free( fullpath, ( path != NULL ) );
	
     	gdk_threads_leave();
}

static void screem_tag_file_parse_insert( ScreemTagFilePrivate *priv,
					const gchar **attributes,
					const gchar **values )
{
	ScreemTagTreeParserState state;
	GSList *top;
	GtkTreeIter *it;
	NodeInfo *info;
	gint i;

	state = GPOINTER_TO_INT( priv->states->data );
	top = priv->iters;

      	gdk_threads_enter();
	
	it = top->data;
	info = NULL;
	if( it ) {
		gtk_tree_model_get( GTK_TREE_MODEL( priv->model ),
					  it, FILE_BROWSER_DATA_COL,
					  &info, -1 );
	}
	if( info ) {
		info->isfunc = ( state == PARSER_STATE_FUNCTION );
		for( i = 0; attributes[ i ]; ++ i ) {
			const gchar *attr;
			const gchar *val;
		
			attr = attributes[ i ];
			val = values[ i ];
		
			if( ! strcmp( "href", attr ) ) {
				GdkPixbuf *pixbuf;	
				gchar *path;
				path = relative_to_full( val, 
							 RESOURCE_PATH );
				info->file = path;
				pixbuf = screem_file_browser_get_icon( priv->browser,
									info->file,
									16, 16, FALSE );
				gtk_tree_store_set( GTK_TREE_STORE( priv->model ), 
						    it, FILE_BROWSER_ICON_COL,
						    pixbuf, -1 );
				g_object_unref( pixbuf );
			} else if( ! strcmp( "open", attr ) ) {
				info->open = g_strdup( val );
			} else if( ! strcmp( "close", attr ) ) {
				info->close = g_strdup( val );
			}
		}
	}
      	gdk_threads_leave();
}


static void screem_tag_file_parse_param( ScreemTagFilePrivate *priv,
					 const gchar **attributes,
					 const gchar **values )
{
	NodeAttribute *attr;
	gint i;
	gint j;
	const gchar *attribute;
	const gchar *value;
	GSList *top;
	GtkTreeIter *it;
	NodeInfo *info;

	top = priv->iters;
	g_assert( top );
	
	it = top->data;
	gdk_threads_enter();
	info = NULL;
	if( it ) {
		gtk_tree_model_get( GTK_TREE_MODEL( priv->model ),
				it, FILE_BROWSER_DATA_COL,
				&info, -1 );
	}
	gdk_threads_leave();

	if( ! info ) {
		g_warning( "<param> with no current node\n" );
		return;
	}
	
	attr = g_new0( NodeAttribute, 1 );
	
	for( i = 0, j = 0; attributes[ i ]; ++ i ) {
		attribute = attributes[ i ];
		
		value = NULL;
		if( values[ j ] ) {
			value = values[ j ];
			++ j;

			if( ! strcmp( "name", attribute ) ) {
				attr->name = g_strdup( value );
				attr->name = g_strstrip( attr->name );
			} else if( ! strcmp( "required", attribute ) ) {
				attr->required = ( *value == '1');
			} else if( ! strcmp( "vallist", attribute ) ) {
				attr->vallist = ( *value == '1' );
			} else if( ! strcmp( "type", attribute ) ) {
				attr->type = g_strdup( value );
			} else if( ! strcmp( "default", attribute ) ) {
				attr->defvalue = g_strdup( value );
				attr->defvalue = g_strstrip( attr->defvalue );
			}
		}
	}
	info->attributes = g_slist_prepend( info->attributes, attr );
}

static gboolean screem_tag_file_autocomplete_fill( GtkTreeModel *model,
						GtkTreePath *path,
						GtkTreeIter *iter,
						gpointer data )
{
	GSList **ret;
	ScreemTagFile *tfile;
	NodeInfo *info;
	const gchar *str;
	const gchar *prefix;

	tfile = SCREEM_TAGFILE( g_object_get_data( G_OBJECT( model ),
						"tagfile" ) );
	prefix = tfile->priv->prefix;
	ret = (GSList**)data;

	gtk_tree_model_get( model, iter,
			FILE_BROWSER_DATA_COL, &info,
			-1 );

	str = info->open;
	if( ! str ) {
		str = info->close;
	}
	if( ! str ) {
		str = info->text;
	}
	if( str && 
	   ! strncmp( str, prefix, strlen( prefix ) ) ) {
		*ret = g_slist_prepend( *ret, (gchar*)str );
	}

	return FALSE;
}

/* public stuff */

ScreemTagFile *screem_tag_file_new( void )
{
	ScreemTagFile *tag_file;

	tag_file = g_object_new( SCREEM_TYPE_TAGFILE, NULL );

	return tag_file;
}

gboolean screem_tag_file_load( ScreemTagFile *tagfile )
{
	GString *file;
	gboolean ret;
	
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), FALSE );
	
	screem_tag_file_clear( tagfile );

	ret = FALSE;
	file = load_file( tagfile->priv->filename );
	if( file ) {
		ret = screem_tag_file_set_from_string( tagfile, 
						file->str,
						file->len );
		g_string_free( file, TRUE );
	}
	
	return ret;	
}

gboolean screem_tag_file_load_with_uri( ScreemTagFile *tagfile,
				const gchar *uri )
{
	screem_tag_file_set_uri( tagfile, uri );

	return screem_tag_file_load( tagfile );
}

gboolean screem_tag_file_set_from_string( ScreemTagFile *tagfile,
					const gchar *str,
					guint length )
{
	ScreemTagFilePrivate *priv;
	GMarkupParseContext *ctx;
	gboolean ret;

	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), FALSE );
	g_return_val_if_fail( str != NULL, FALSE );
	
	priv = tagfile->priv;
	
	if( length == 0 ) {
		length = strlen( str );
	}
	
	ret = FALSE;

	priv->states = g_slist_prepend( NULL,
					GINT_TO_POINTER( PARSER_STATE_NONE ) );
	priv->iters = g_slist_prepend( priv->iters, NULL );
	ctx = g_markup_parse_context_new( &parser_cbs, 0, 
					  priv, NULL );
	if( ctx ) {
		ret = g_markup_parse_context_parse( ctx, str, 
							length, NULL );
		g_markup_parse_context_free( ctx );
	}
	if( priv->iters->next ) {
		g_slist_foreach( priv->iters,
				 (GFunc)gtk_tree_iter_free, NULL );
	}
	g_slist_free( priv->iters );
	g_slist_free( priv->states );

	priv->iters = NULL;
	priv->states = NULL;

	priv->loaded = ret;
	
	return ret;	
}

GtkTreeModel *screem_tag_file_get_model( ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->model;
}

const gchar *screem_tag_file_get_name( const ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->name;
}

const gchar *screem_tag_file_get_desc( const ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->description;
}

const gchar *screem_tag_file_get_uri( const ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->filename;
}

const gchar *screem_tag_file_for_type( const ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->type;
}

void screem_tag_file_set_name( ScreemTagFile *tagfile, 
				const gchar *name )
{
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );
	g_return_if_fail( name != NULL );

	g_free( tagfile->priv->name );
	tagfile->priv->name = g_strdup( name );
}

void screem_tag_file_set_desc( ScreemTagFile *tagfile,
				const gchar *desc )
{
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );
	g_return_if_fail( desc != NULL  );

	g_free( tagfile->priv->description );
	tagfile->priv->description = g_strdup( desc );
}

void screem_tag_file_set_uri( ScreemTagFile *tagfile,
				const gchar *uri )
{
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );
	g_return_if_fail( uri != NULL );

	g_free( tagfile->priv->filename );
	tagfile->priv->filename = g_strdup( uri );
}

GSList *screem_tag_file_autocomplete( ScreemTagFile *tagfile,
					const gchar *prefix )
{
	ScreemTagFilePrivate *priv;
	GSList *ret;
	
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );
	g_return_val_if_fail( prefix != NULL, NULL );

	priv = tagfile->priv;
	
	ret = NULL;
	priv->prefix = prefix;
	gtk_tree_model_foreach( priv->model,
			(GtkTreeModelForeachFunc)screem_tag_file_autocomplete_fill,
			&ret );

	return ret;			
}

