/*  Screem:  screem-tree-view.c
 *
 *  The tree view widget
 *
 *  Copyright (C) 2002  David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <string.h>

#include <libgnome/gnome-i18n.h>

#include <gtk/gtktreestore.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcellrenderertoggle.h>
#include <gtk/gtkcellrendererpixbuf.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkstock.h>

#include "screem-tree-view.h"

#include "screem-window.h"

#include "screem-markup.h"
#include "screem-dtd.h"

#include "screemmarshal.h"

enum {
	TREE_VIEW_ICON_COL = 0,
	TREE_VIEW_NAME_COL,
	TREE_VIEW_TEXT_COL,
	TREE_VIEW_START_COL,
	TREE_VIEW_END_COL,
	TREE_VIEW_VALID_COL,
	TREE_VIEW_MAX_COL
};

enum {
	SELECT_CONTEXT,
	LAST_SIGNAL
};

static guint screem_tree_view_signals[ LAST_SIGNAL ] = { 0 };

typedef struct {
	GtkTreeModel *model;
	GtkTreeIter document;

	GSList *stack;
	
	GtkTreeIter *parent;
	gchar *pname;
	ScreemDTDTagExistance pclose;

	guint pos;
	guint ppos;
	gchar *text;
	gchar *name;
	gchar *tag;
	guint start;
	guint end;
	guint last;

	ScreemDTD *dtd;
	gboolean gotroot;

	guint sourceid;
	ScreemTreeView *view;
} ScreemTreeViewParser;

static void screem_tree_view_class_init( ScreemTreeViewClass *klass );
static void screem_tree_view_init( ScreemTreeView *tree_view );
static void screem_tree_view_finalize( GObject *tree_view );
static void screem_tree_view_window_set( ScreemView *view );
static void screem_tree_view_size_request( GtkWidget *widget, 
					   GtkRequisition *req );
static void screem_tree_view_realize( GtkWidget *widget );
static void screem_tree_view_show( GtkWidget *widget );
static void screem_tree_view_hide( GtkWidget *widget );

static void screem_tree_view_display( ScreemView *view );
static void screem_tree_view_page( ScreemTreeView *view, ScreemPage *page );

static gboolean select_context( ScreemTreeView *view,
				GtkTreeModel *model, GtkTreeIter **it,
				guint pos, guint *start, guint *end,
				gboolean select_text,
				gboolean build );
static void page_change( ScreemPage *page, GParamSpec *spec, 
			 gpointer data );
static gboolean build_tree( ScreemTreeView *view );

static gboolean screem_tree_view_clear_tree( GtkTreeModel *model, 
					     GtkTreePath *path, 
					     GtkTreeIter *iter, 
					     gpointer data );

static void screem_tree_view_parser_cleanup( ScreemTreeViewParser *parser );
static gboolean screem_tree_view_build_step( ScreemTreeViewParser *parser );

static void screem_tree_view_create_mini_view( ScreemTreeView *view );
static void screem_tree_view_mini_activate( GtkTreeView *mini,
					    GtkTreePath *path,
					    GtkTreeViewColumn *column,
					    gpointer data );
struct ScreemTreeViewPrivate {
	GtkTreeModel *model;
	GtkTreeView *view;

	GtkTreeView *mini;

	guint timeout;
	ScreemPage *current;

	GtkTreeIter *context;

	ScreemTreeViewParser parser;

	GtkWidget *sw;
};

ScreemTreeView *screem_tree_view_new( ScreemWindow *window )
{
	ScreemTreeView *view;
	GType type;

	type = screem_tree_view_get_type();

	view = SCREEM_TREE_VIEW( g_object_new( type, "window", window, NULL) );

	return view;
}

gboolean screem_tree_view_select_element_context( ScreemTreeView *view,
						  guint pos,
						  guint *start, guint *end,
						  gboolean select_text )
{
	GtkTreeModel *model;
	guint tstart;
	guint tend;
	gboolean ret;

	g_return_val_if_fail( SCREEM_IS_TREE_VIEW( view ), FALSE );

	model = GTK_TREE_MODEL( view->private->model );

	if( ! start ) {
		start = &tstart;
	}
	if( ! end ) {
		end = &tend;
	}

	ret = select_context( view, 
			      model, NULL, pos, start, end, select_text, TRUE );
	if( ret ) {
		/* emit signal */
		g_signal_emit( G_OBJECT( view ),
			       screem_tree_view_signals[ SELECT_CONTEXT ],
			       0, *start, (*end) );
	}

	return ret;
}

gchar *screem_tree_view_query_element_context( ScreemTreeView *view,
					       guint pos,
					       gboolean query_text,
					       gboolean build,
					       guint *depth,
					       guint *start,
					       guint *end )
{
	GtkTreeModel *model;
	guint s;
	guint e;
	gboolean ret;
	gchar *retval;

	g_return_val_if_fail( SCREEM_IS_TREE_VIEW( view ), FALSE );

	model = GTK_TREE_MODEL( view->private->model );

	if( ! start ) {
		start = &s;
	}
	if( ! end ) {
		end = &e;
	}

	ret = select_context( view, 
			      model, NULL, pos, start, end,
			      query_text, build );
	retval = NULL;
	/* FIXME: view->private->context should always be non NULL
	   if ret is TRUE, however we don't handle <!DOCTYPE ... [ .. ]>
	   very well, and can cause it to be NULL, at least I think
	   that is what is causing a crash */
	if( ret && view->private->context ) {
		GtkTreePath *path;

		gtk_tree_model_get( model, view->private->context,
				TREE_VIEW_NAME_COL, &retval, -1 );

		path = gtk_tree_model_get_path( model, 
					view->private->context );

		if( depth ) {
			/* -1 as we are always under the [document] node */
			*depth = gtk_tree_path_get_depth( path );
			if( *depth > 0 ) {
				(*depth) --;
			}
		}
		gtk_tree_path_free( path );

		gtk_tree_iter_free( view->private->context );
		view->private->context = NULL;

	} else if( depth ) {
		*depth = 0;
	}

	return retval;
}

/* static stuff */
static void screem_tree_view_display( ScreemView *view )
{
	ScreemPage *page;

	g_object_get( G_OBJECT( view ), "page", &page, NULL );

	screem_tree_view_page( SCREEM_TREE_VIEW( view ), page );

	if( page ) {
		g_object_unref( page );
	}
}

static void screem_tree_view_page( ScreemTreeView *view, ScreemPage *page )
{
	if( view->private->current == page ) {
		return;
	}

	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( view->private->sw ),
					GTK_POLICY_AUTOMATIC,
                        	        GTK_POLICY_AUTOMATIC );

	if( view->private->timeout ) {
		g_source_remove( view->private->timeout );
		view->private->timeout = 0;
	}
	
	if( view->private->current ) {
		g_signal_handlers_disconnect_matched( G_OBJECT( view->private->current ),
						      G_SIGNAL_MATCH_DATA,
						      0, 0, NULL, NULL, view );
	}
	
	view->private->current = page;
	if( page ) {
		page_change( page, NULL, view );
		g_signal_connect( G_OBJECT( page ), "notify::changed",
				  G_CALLBACK( page_change ), view );
		g_signal_connect( G_OBJECT( page ), "notify::dtd",
				  G_CALLBACK( page_change ), view );
	}
}

static gboolean select_context( ScreemTreeView *view,
				GtkTreeModel *model, GtkTreeIter **it,
				guint pos,
				guint *start, guint *end,
				gboolean select_text,
				gboolean build )
{
	ScreemTreeViewPrivate *priv;
	gboolean loop;
	GtkTreeIter *tmp;
	GtkTreeIter root;
	
	priv = view->private;
	
	/* If the build timer is going, remove it and construct a new tree,
	   but only if a build was requested */
	gdk_threads_leave();
	if( build && priv->timeout ) {
		g_source_remove( priv->timeout );
		priv->timeout = 0;
		build_tree( view );
	}
	/* make sure full tree is fully built, we remove
	   the idle handler, but don't set the handle to 0
	   as this will be done by while building in the
	   while loop */
	if( priv->parser.sourceid ) {
		g_source_remove( priv->parser.sourceid );
	}
	while( priv->parser.sourceid ) {
		screem_tree_view_build_step( &priv->parser );
	}
	gdk_threads_enter();

	loop = TRUE;
	if( ! it ) {
		if( priv->context ) {
			gtk_tree_iter_free( priv->context );
			priv->context = NULL;
		}

		loop = gtk_tree_model_get_iter_first( model, &root );
		if( loop && gtk_tree_model_iter_has_child( model, &root ) ) {
			GtkTreeIter *parent;
			
			parent = gtk_tree_iter_copy( &root );

			loop = gtk_tree_model_iter_children( model, &root, 
							     parent );
			if( loop ) {
				tmp = &root;
				it = &tmp;
			}
			gtk_tree_iter_free( parent );
		} else {
			loop = FALSE;
		}
	}
	
	while( loop ) {
		gint estart;
		gint eend;

		gtk_tree_model_get( model, *it, 
				TREE_VIEW_START_COL, &estart,
				TREE_VIEW_END_COL, &eend, -1 );

		if( pos <= eend && pos >= estart ) {
			/* in here somewhere */
			gboolean istext;
			GValue value = { 0 };

			gtk_tree_model_get_value( model, *it,
						  TREE_VIEW_TEXT_COL,
						  &value );
			istext = ( g_value_get_string( &value ) != NULL );
			g_value_unset( &value );

			if( ! istext || select_text ) {
				*start = estart;
				*end = eend + 1;
				if( view->private->context ) {
					gtk_tree_iter_free( view->private->context );
				}
				view->private->context = 
					gtk_tree_iter_copy( *it );
			}
			
			if( gtk_tree_model_iter_has_child( model, *it ) ) {
				GtkTreeIter child;
				
				if( gtk_tree_model_iter_children( model,
								  &child,
								  *it ) ) {
					GtkTreeIter *cptr;
					
					cptr = &child;
					select_context( view, 
							model, &cptr, pos,
							start, end,
							select_text,
							FALSE );
				}
			}
			break;
		} else {
			loop = gtk_tree_model_iter_next( model, *it );
		}
	}

	return loop;
}

static void page_change( ScreemPage *page, GParamSpec *spec, 
			 gpointer data )
{
	ScreemTreeView *view;
	gboolean status;

	view = SCREEM_TREE_VIEW( data );
	status = screem_page_get_changed( page );
	status |= ( spec == NULL );
	
	if( view->private->timeout ) {
		g_source_remove( view->private->timeout );
	}
#if 0
	view->private->timeout = g_idle_add_full( G_PRIORITY_LOW,
				(GSourceFunc)build_tree, view, NULL );
#else
	view->private->timeout = g_timeout_add( 1000, 
						(GSourceFunc)build_tree,
						view );
#endif
}

static gboolean build_tree( ScreemTreeView *view )
{
	ScreemTreeViewPrivate *priv;
	ScreemPage *page;
	GtkTreeModel *model;
	gboolean feature_markup;
	ScreemTreeViewParser *parser;
	
	priv = view->private;
	
	page = priv->current;
	if( page ) {
		gdk_threads_enter();
	
		model = GTK_TREE_MODEL( view->private->model );

		/* this will stop any current build + 
		   cleanup after it */
		screem_tree_view_parser_cleanup( &priv->parser );

		gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( priv->sw ),
						GTK_POLICY_ALWAYS,
                	                        GTK_POLICY_ALWAYS );

		gtk_tree_view_set_model( priv->view, NULL );
		gtk_tree_view_set_model( priv->mini, NULL );

		gtk_tree_model_foreach( model, screem_tree_view_clear_tree, NULL );
		gtk_tree_store_clear( GTK_TREE_STORE( model ) );
	
		feature_markup = screem_page_is_markup( page );

		if( feature_markup ) {
			/* init parser */
			parser = &view->private->parser;
			parser->dtd = screem_page_get_dtd( page );
			gtk_tree_store_append( GTK_TREE_STORE( parser->model ),
					&parser->document, NULL );
			gtk_tree_store_set( GTK_TREE_STORE( parser->model ), 
				    &parser->document,
				    TREE_VIEW_NAME_COL,
				    _( "[Document]" ),
				    TREE_VIEW_VALID_COL, TRUE,
				    -1 );
			parser->parent = &parser->document;
			parser->pclose = SCREEM_DTD_MUST;
			parser->text = screem_page_get_data( page );
			parser->last = gtk_text_buffer_get_char_count( GTK_TEXT_BUFFER( page ) ); 

			/* add parser idle handler */
			parser->sourceid = g_idle_add_full( G_PRIORITY_LOW,
				(GSourceFunc)screem_tree_view_build_step,
				parser, NULL );
		} else {
			gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( view->private->sw ),
						GTK_POLICY_AUTOMATIC,
                        	                GTK_POLICY_AUTOMATIC );
		}

		view->private->timeout = 0;
		
		gdk_threads_leave();
	}
	
	return FALSE;
}

static gboolean screem_tree_view_clear_tree( GtkTreeModel *model, 
					     GtkTreePath *path,
					     GtkTreeIter *iter, 
					     gpointer data )
{

	return FALSE;
}

static void screem_tree_view_build_text_node( ScreemTreeViewParser *parser )
{
	/* add text node */
	const gchar *offset;
	GtkTreeIter tit;
	gchar *txt;
	gboolean valid;

	offset = g_utf8_offset_to_pointer( parser->text,
					parser->ppos );
	txt = g_strndup( offset, parser->pos - parser->ppos );
		
	/* chomp / chop any white space, if all 
	   white space then we won't even bother
	   putting it in the tree */
	txt = g_strstrip( txt );

	if( *txt != '\0' ) {
		valid = screem_dtd_valid_child( parser->dtd, 
				parser->pname, "PCDATA" );
		if( ! valid ) {
			valid = screem_dtd_valid_child( parser->dtd,
						     parser->pname,
						     "CDATA");
		}

		gtk_tree_store_append( GTK_TREE_STORE( parser->model ),
				       &tit, parser->parent );
		gtk_tree_store_set( GTK_TREE_STORE( parser->model ), 
				    &tit,
				    TREE_VIEW_NAME_COL,
				    _( "[text]" ),
				    TREE_VIEW_TEXT_COL,
				    txt,
				    TREE_VIEW_START_COL,
				    parser->ppos,
				    TREE_VIEW_END_COL,
				    parser->pos,
				    TREE_VIEW_VALID_COL,
				    valid,
				    -1 );
	}
	g_free( txt );
}

static void screem_tree_view_build_close( ScreemTreeViewParser *parser )
{
	gboolean match;

	/* pop stack, change parent until
	   we hit the correct close tag or
	   we run out of stack */
	match = FALSE;
	g_free( parser->pname );
	while( parser->stack && ! match ) {
		gtk_tree_store_set( GTK_TREE_STORE( parser->model ),
				    parser->parent,
				   TREE_VIEW_END_COL, 
				   parser->pos - 1, -1 );

		parser->stack = g_slist_remove( parser->stack, 
					parser->stack->data );
		/* get parent name */
		gtk_tree_model_get( parser->model, parser->parent,
				  TREE_VIEW_NAME_COL,
				  &parser->pname, -1 );
		match = ! strcmp( parser->name + 1, parser->pname );
		g_free( parser->pname );
		gtk_tree_iter_free( parser->parent );
		if( parser->stack ) {
			parser->parent = parser->stack->data;
		}
	}

	parser->pname = NULL;
	if( ! parser->stack ) {
		parser->parent = &parser->document;
		parser->pclose = SCREEM_DTD_MUST;
	} else if( parser->parent != &parser->document ) {
		gtk_tree_model_get( parser->model,
				  parser->parent,
				  TREE_VIEW_NAME_COL,
				  &parser->pname, -1 );
		parser->pclose = screem_dtd_element_close_state( parser->dtd,
						 parser->pname );
	}
}

static void screem_tree_view_build_open( ScreemTreeViewParser *parser )
{
	gchar firstc;
	gboolean valid;
	gboolean iscomment;
	ScreemDTDTagExistance close_state;
	GtkTreeIter it;
	GSList *attrs;
	GSList *tmp;
	gchar *name;
	gchar *value;
	
	firstc = *parser->name;
	
	if( ( iscomment = ! strcmp( "!--", parser->name ) ) ) {
		g_free( parser->name );
		parser->name = g_strdup( _( "[Comment]" ) );
	}

	close_state = screem_dtd_element_close_state( parser->dtd,
						parser->name );

	if( firstc == '!' || firstc == '?' ) {
		valid = TRUE;
	} else if( parser->parent == &parser->document &&
		! parser->gotroot &&
		screem_dtd_is_root_element( parser->dtd, parser->name )) {

		valid = TRUE;
		parser->gotroot = TRUE;
	} else {
		valid = screem_dtd_valid_child( parser->dtd,
					parser->pname,
					parser->name );
	}
			
	/* if closing the parent tag is optional, and 
	   valid is FALSE, pop the parent */
	while( parser->parent != &parser->document && 
		parser->pclose == SCREEM_DTD_SHOULD &&
	       ! valid ) {
		/* assume this closes the parent */
		gtk_tree_store_set( GTK_TREE_STORE( parser->model ),
				parser->parent, TREE_VIEW_END_COL, 
				parser->start, -1 );
		parser->stack = g_slist_remove( parser->stack, 
						parser->stack->data );
		gtk_tree_iter_free( parser->parent );
		if( parser->stack ) {
			parser->parent = parser->stack->data;
				
			g_free( parser->pname );
			gtk_tree_model_get( parser->model, 
					parser->parent,
					TREE_VIEW_NAME_COL,
					&parser->pname, -1 );
			parser->pclose = screem_dtd_element_close_state( parser->dtd, parser->pname );
			valid = screem_dtd_valid_child( parser->dtd,
					parser->pname,
					parser->name );
		} else {
			parser->parent = &parser->document;
			g_free( parser->pname );
			parser->pname = NULL;
		}
	}
			
	/* we are dealing with an opening tag */
	gtk_tree_store_append( GTK_TREE_STORE( parser->model ),
			       &it, parser->parent );
	if( iscomment ) {
		gtk_tree_store_set( GTK_TREE_STORE( parser->model ),
				&it, TREE_VIEW_TEXT_COL,
				parser->tag,
				-1 );
	}

	/* set values for the iter */
	gtk_tree_store_set( GTK_TREE_STORE( parser->model ),
			&it,
			TREE_VIEW_NAME_COL, parser->name,
			TREE_VIEW_START_COL,parser->start,
			TREE_VIEW_END_COL, parser->end,
			TREE_VIEW_VALID_COL, valid,
			-1 );

	/* TODO:
	 * check attributes + validate
	 * */
#if 0
	attrs = screem_markup_build_attributes_list( parser->tag );
	attrs = g_slist_reverse( attrs );
	for( tmp = attrs; tmp; tmp = tmp->next ) {
		name = tmp->data;
		tmp = tmp->next;
		value = tmp->data;
	
		valid = screem_dtd_valid_attr( parser->dtd,
						parser->name,
						name );
		g_free( name );
		g_free( value );
	}
	g_slist_free( attrs );
#endif

	/* put it on the stack as a parent if needed,
	   we don't need to for:
	   1) comments / doctypes
	   2) tags closing themselves, eg <meta /> 
	   3) HTML tags which forbid closing, same as
	      2 but we need to lookup in the doctype */
	if(  ( firstc != '!' ) && ( firstc != '?' ) &&
	    ( parser->text[ parser->end - 1 ] != '/' ) &&
		( close_state != SCREEM_DTD_MUST_NOT ) ) {
		parser->parent = gtk_tree_iter_copy( &it );
		parser->stack = g_slist_prepend( parser->stack, 
						parser->parent );
		g_free( parser->pname );
		gtk_tree_model_get( parser->model,
				  parser->parent,
				  TREE_VIEW_NAME_COL,
				  &parser->pname, -1 );
		parser->pclose = screem_dtd_element_close_state( parser->dtd, parser->pname );
	}
}

static void screem_tree_view_parser_cleanup( ScreemTreeViewParser *parser )
{
	GSList *cleanup;
	GtkTreeModel *model;
	ScreemTreeView *view;

	model = parser->model;
	view = parser->view;
	
	g_free( parser->pname );
	
	/* cleanup the stack */
	cleanup = parser->stack;
	while( cleanup ) {
		gtk_tree_store_set( GTK_TREE_STORE( parser->model ),
				   parser->parent,
				   TREE_VIEW_END_COL, 
				   parser->pos, -1 );
		cleanup = g_slist_remove( cleanup, cleanup->data ); 
		if( parser->parent != &parser->document ) {
			gtk_tree_iter_free( parser->parent );
		}
		if( cleanup ) {
			parser->parent = cleanup->data;
		}
	}
	g_free( parser->text );

	if( parser->sourceid ) {
		g_source_remove( parser->sourceid );
		parser->sourceid = 0;
	}

	memset( parser, 0, sizeof( ScreemTreeViewParser ) );

	parser->model = model;
	parser->view = view;
}

static gboolean screem_tree_view_build_step( ScreemTreeViewParser *parser )
{
	gboolean ret;
	ScreemTreeView *view;
	ScreemTreeViewPrivate *priv;
	GtkWidget *widget;
	
	ret = TRUE;

	parser->ppos = parser->pos;

	parser->name = NULL;
		
	parser->tag = screem_markup_next_tag( parser->text, 
				parser->pos, 
				&parser->start, &parser->end, 
				&parser->name );
	if( ! parser->tag ) {
		parser->pos = parser->last;
	} else {
		parser->pos = parser->start;
	}

	gdk_threads_enter();

	if( parser->pos > parser->ppos ) {
		/* TODO?
		 *
		 * If we are in a <script> we could possibly
		 * split the text node based on blocks, 
		 * e.g. { } 
		 *
		 * Likewise if we are in <style> we could
		 * split out the individual CSS classes etc
		 */
		screem_tree_view_build_text_node( parser );
	}
		
	if( parser->tag ) {
		parser->pos = parser->end + 1;
		if( *parser->name == '/' ) {
			screem_tree_view_build_close( parser );
		} else {
			screem_tree_view_build_open( parser );
		}
	}
	g_free( parser->name );
	g_free( parser->tag );

	if( parser->pos >= parser->last ) {
		ret = FALSE;
		parser->sourceid = 0;
		screem_tree_view_parser_cleanup( parser );

		view = parser->view;
		priv = view->private;

		gtk_tree_view_set_model( priv->view, priv->model );
		gtk_tree_view_set_model( priv->mini, priv->model );
	
		gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( priv->sw ),
						GTK_POLICY_AUTOMATIC,
                        	                GTK_POLICY_AUTOMATIC );

		widget = GTK_WIDGET( priv->mini );
		gtk_tree_view_expand_all( GTK_TREE_VIEW( widget ) );

		widget = GTK_WIDGET( priv->view );
		if( GTK_WIDGET_VISIBLE( widget ) ) {
			gtk_tree_view_expand_all( GTK_TREE_VIEW( widget ) );
		}
	}

	gdk_threads_leave();

	return ret;
}

static void screem_tree_view_create_mini_view( ScreemTreeView *view )
{
	ScreemTreeViewPrivate *private;
	GtkWidget *box;
	GtkWidget *tree;
	GtkWidget *sw;
	GtkCellRenderer *rend;
	GtkCellRenderer *prend;
	GtkTreeViewColumn *col;
	ScreemWindow *window;
	
	private = view->private;

	box = gtk_vbox_new( FALSE, 0 );
	gtk_widget_show( box );

	tree = gtk_tree_view_new_with_model( GTK_TREE_MODEL(private->model) );

	private->mini = GTK_TREE_VIEW( tree );
	gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( tree ), TRUE );

	private->sw = sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ),
					GTK_POLICY_AUTOMATIC,
                                        GTK_POLICY_AUTOMATIC );
	gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
						GTK_SHADOW_IN );
	prend = gtk_cell_renderer_pixbuf_new();
	rend = gtk_cell_renderer_text_new ();
	col = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title( col, _( "Element" ) );
	gtk_tree_view_column_pack_start( col, prend, FALSE );
	gtk_tree_view_column_pack_start( col, rend, TRUE );
	gtk_tree_view_column_set_attributes( col, rend,
					     "text", TREE_VIEW_NAME_COL,
					     NULL );
	gtk_tree_view_column_set_attributes( col, prend,
					     "pixbuf",
					     TREE_VIEW_ICON_COL,
					     NULL );
	gtk_tree_view_column_set_resizable( col, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), col );

	rend = gtk_cell_renderer_toggle_new();
	col = gtk_tree_view_column_new_with_attributes( _( "Valid" ),
							rend,
							"active",
							TREE_VIEW_VALID_COL,
							NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), col );
	
	gtk_container_add( GTK_CONTAINER( sw ), tree );
	gtk_widget_show( GTK_WIDGET( tree ) );
	gtk_widget_show( sw );

	gtk_box_pack_start( GTK_BOX( box ), sw, TRUE, TRUE, 0 );

	g_object_get( G_OBJECT( view ), "window", &window, NULL );
	screem_window_add_dock_item( window, _( "Tree" ), 
				     GTK_STOCK_INDEX, 
				     "Structure",
				     box, GTK_POS_RIGHT );

	
	g_signal_connect( G_OBJECT( tree ), "row_activated",
			  G_CALLBACK( screem_tree_view_mini_activate ),
			  view );
}

static void screem_tree_view_mini_activate( GtkTreeView *mini, 
					    GtkTreePath *path,
					    GtkTreeViewColumn *column,
					    gpointer data )
{
	ScreemTreeView *view;
	GtkTreeModel *model;
	GtkTreeIter it;
	guint start;
	guint end;

	view = SCREEM_TREE_VIEW( data );

	model = gtk_tree_view_get_model( mini );
	gtk_tree_model_get_iter( GTK_TREE_MODEL( model ), &it, path );

	gtk_tree_model_get( model, &it, 
			TREE_VIEW_START_COL, &start,
			TREE_VIEW_END_COL, &end, -1 );

	/* emit signal */
	if( start != end ) {
		g_signal_emit( G_OBJECT( view ),
			       screem_tree_view_signals[ SELECT_CONTEXT ],
			       0, start, end  + 1 );
	}
}

/* G Object stuff */
#define PARENT_TYPE SCREEM_TYPE_VIEW

static gpointer parent_class;

static void screem_tree_view_class_init( ScreemTreeViewClass *klass )
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	object_class = G_OBJECT_CLASS( klass );
	widget_class = (GtkWidgetClass *)klass;
	parent_class = g_type_class_peek_parent( klass );

	object_class->finalize = screem_tree_view_finalize;

	widget_class->realize = screem_tree_view_realize;
	widget_class->size_request = screem_tree_view_size_request;
	widget_class->show = screem_tree_view_show;
	widget_class->hide = screem_tree_view_hide;

	screem_tree_view_signals[ SELECT_CONTEXT ] = 
		g_signal_new( "select_context",
			      G_OBJECT_CLASS_TYPE( object_class ),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET( ScreemTreeViewClass, 
					       select_context),
			      NULL, NULL,
			      screem_marshal_VOID__UINT_UINT,
			      G_TYPE_NONE, 2,
			      G_TYPE_UINT,
			      G_TYPE_UINT );
}

static void screem_tree_view_init( ScreemTreeView *tree_view )
{
	ScreemTreeViewPrivate *private;
	GtkWidget *sw;
	GtkCellRenderer *rend;
	GtkCellRenderer *prend;

	GtkTreeViewColumn *col;

	SCREEM_VIEW( tree_view )->window_set = screem_tree_view_window_set;
	SCREEM_VIEW( tree_view )->display = screem_tree_view_display;

	private = tree_view->private = g_new0( ScreemTreeViewPrivate, 1 );

	sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ),
					GTK_POLICY_AUTOMATIC,
                                        GTK_POLICY_AUTOMATIC );
	gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
						GTK_SHADOW_IN );

	private->model = GTK_TREE_MODEL( gtk_tree_store_new( TREE_VIEW_MAX_COL,
							     GDK_TYPE_PIXBUF,
							     G_TYPE_STRING,
							     G_TYPE_STRING,
							     G_TYPE_UINT,
							     G_TYPE_UINT,
							     G_TYPE_BOOLEAN) );
	private->view = GTK_TREE_VIEW(
				      gtk_tree_view_new_with_model( GTK_TREE_MODEL(private->model) ) );
	gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( private->view ), TRUE );

	prend = gtk_cell_renderer_pixbuf_new();
	rend = gtk_cell_renderer_text_new ();
	col = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title( col, _( "Element" ) );
	gtk_tree_view_column_pack_start( col, prend, FALSE );
	gtk_tree_view_column_pack_start( col, rend, TRUE );
	gtk_tree_view_column_set_attributes( col, rend,
					     "text", TREE_VIEW_NAME_COL,
					     NULL );
	gtk_tree_view_column_set_attributes( col, prend,
					     "pixbuf",
					     TREE_VIEW_ICON_COL,
					     NULL );
	gtk_tree_view_column_set_resizable( col, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( private->view ), col );

	rend = gtk_cell_renderer_text_new ();
	col = gtk_tree_view_column_new_with_attributes( _( "Text" ),
							rend,
							"text", 
							TREE_VIEW_TEXT_COL,
							NULL );
	gtk_tree_view_column_set_resizable( col, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( private->view ), col );

	rend = gtk_cell_renderer_toggle_new();
	col = gtk_tree_view_column_new_with_attributes( _( "Valid" ),
							rend,
							"active",
							TREE_VIEW_VALID_COL,
							NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW( private->view ), col );
	
	gtk_container_add( GTK_CONTAINER( sw ), GTK_WIDGET( private->view ) );
	gtk_container_add( GTK_CONTAINER( tree_view ), sw );

	gtk_widget_show( GTK_WIDGET( private->view ) );
	gtk_widget_show( sw );

	private->timeout = 0;
	private->parser.view = tree_view;
	private->parser.model = GTK_TREE_MODEL( private->model );	

}

static void screem_tree_view_finalize( GObject *tree_view )
{
	ScreemTreeView *view;

	view = SCREEM_TREE_VIEW( tree_view );

	screem_tree_view_parser_cleanup( &view->private->parser );

	if( view->private->timeout ) {
		g_source_remove( view->private->timeout );
	}

	g_free( view->private );

	G_OBJECT_CLASS( parent_class )->finalize( tree_view );
}

static void screem_tree_view_window_set( ScreemView *view )
{
	/* add miniview tab */
	screem_tree_view_create_mini_view( SCREEM_TREE_VIEW( view ) );
}

static void screem_tree_view_size_request( GtkWidget *widget, 
					   GtkRequisition *req )
{
	GTK_WIDGET_CLASS( parent_class )->size_request( widget, req );
}

static void screem_tree_view_realize( GtkWidget *widget )
{
	GTK_WIDGET_CLASS( parent_class )->realize( widget );
}

static void screem_tree_view_show( GtkWidget *widget )
{
	ScreemTreeView *view;

	view = SCREEM_TREE_VIEW( widget );

	GTK_WIDGET_CLASS( parent_class )->show( widget );

	gtk_widget_set_sensitive( GTK_WIDGET( view->private->mini ), FALSE );
	
	gtk_widget_show( GTK_WIDGET( view->private->view ) );
}

static void screem_tree_view_hide( GtkWidget *widget )
{
	ScreemTreeView *view;

	view = SCREEM_TREE_VIEW( widget );

	GTK_WIDGET_CLASS( parent_class )->hide( widget );

	gtk_widget_set_sensitive( GTK_WIDGET( view->private->mini ), TRUE );
	
	gtk_widget_hide( GTK_WIDGET( view->private->view ) );
}

GType screem_tree_view_get_type()
{
	static guint type = 0;

	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemTreeViewClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_tree_view_class_init,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemTreeView ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_tree_view_init
		};
		
		type = g_type_register_static( PARENT_TYPE,
					       "ScreemTreeView",
					       &info, 0 );
	}

	return type;
}

